==> vmp-0.3.3.tar.xz: vmp-0.3.3/framework/src/event/manager.cc <==
/* -*- Mode:C++; c++-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Marco Guastella alias Vasta 
 * Web page:<www.ragnu.it> 
 * Email:<vasta@ragnu.it>
   Date last update: 15/05/2020
 */

#include "event.h"

namespace vampiria { namespace event {

Manager::Manager()
{
    timer_.init();
    timeout_=0.0;
    now_=0.0;
    loop_=true;
    self_=vmp::thread::pthread_self_wrap();
    breakui_=new event::BreakloopUI(this);
    breakui_->cell_->priority_=event::BREAKLOOP_PRIORITY;
}

Manager::~Manager()
{
    delete breakui_;
    breakui_=0;
}

void Manager::lock()
{
    mutex_.lock();
}

void Manager::unlock()
{
    mutex_.unlock();
}

void Manager::breakloop()
{
    if(self_ != vmp::thread::pthread_self_wrap())
       breakui_->active();       
}

event::Cell *Manager::cell_new(event::UI *ui,event::Event *event,vmp_int priority,event::EVTCB read,event::EVTCB close)
{
    event::Cell *cell=cellref_.get();
    cell->reset();
    cell->ui_=ui;
    cell->event_=event;
    event->cell_=cell;
    cell->manager_=this;
    cell->priority_=priority;
    cell->writing_=false;
    if(read == 0)
      cell->read_=event::empty_ev;
    else
      cell->read_=read;
    if(close == 0)
      cell->close_=event::empty_ev;
    else
      cell->close_=close;
    cell_alloc(cell);
    pcell_.push(cell);
    breakloop();
    return cell;  
}

vmp_bool Manager::cell_update(event::Cell *cell)
{
    if(cell->state_ == event::CLOSE)
       return false;
    if(cell->state_ == event::CLOSEWAIT)
    {
        if(cell->nwait_ == 0)
            return false;
        cell->nwait_--;
    }
    if(cell->timewait_ != 0.0)
       cell->timeout_=now_+cell->timewait_;
    return true;
}

void Manager::cell_timewait(event::Cell *cell,vmp::time::Time timewait)
{
    if(timewait > 0.0)
    {
        cell->timewait_=timewait;
        cell->timeout_=now_+cell->timewait_;
    }
    else
    {
        cell->timewait_=0.0;
        cell->timeout_=0.0;
    }
    breakloop();
}

void Manager::cell_close(event::Cell *cell,event::CELLRET retcode,vmp::str errinfo)
{
   if(cell->state_ != event::CLOSE)
   {
       cell->state_=event::CLOSE;
       cell->ret_=retcode;
       cell->err_=errinfo;
       cell->ui_->close_event(cell);
       breakloop();
   }
}

void Manager::cell_closewait(event::Cell *cell,vmp_uint n)
{
   if(cell->state_ == event::ACTIVE)
   {
       if(n == 0)
           cell_close(cell,event::SUCCESS);
       else
       {
           cell->state_=event::CLOSEWAIT;
           cell->nwait_=n;
       }
   }
}

vmp::time::Time Manager::time_now()
{
    return now_;
}

void Manager::set_timeout(vmp::time::Time timeout)
{
    if(timeout > 0.0)
    {
       if((timeout_ == 0.0) || (timeout < timeout_))
          timeout_=timeout;  
    }
}

void Manager::cell_alloc(event::Cell *cell)
{
    cell->refalloc_++;
}

void Manager::cell_release(event::Cell *cell)
{
    cell->refalloc_--;
    if(cell->refalloc_==0)
    {
        cell->ui_->free_ref(cell);
        cellref_.free(cell);
    }
}

void Manager::stop()
{
    loop_=false;
    breakui_->active();
}

void Manager::loop()
{
    register vmp::Select_FD select;
    register event::Cell *cell;
    register vmp::vector<event::Cell *> vread;
    register vmp::vector<event::Cell *> vclose;
    register vmp::time::Time stimeout;
    while(1)
    {
        lock();
        select.reset();
        vread.clear();
        vclose.clear();
        while(!pcell_.empty())
        {
            cell=pcell_.top();
            pcell_.pop();
            if(((cell->state_ == event::CLOSEWAIT) && (cell->nwait_ == 0)) || (!loop_))
                cell_close(cell,event::SUCCESS);   
            if(cell->state_ == event::CLOSE)
                vclose.push_back(cell);    
            else
            {
                vread.push_back(cell);
                select.fd_add(cell->event_->fd_);
                if(cell->writing_)
                    select.fd_add_w(cell->event_->fd_);
                set_timeout(cell->timeout_);
            }
          
        }
        if(timeout_ > 0.0)
           stimeout=timeout_-now_;
        else
           stimeout=0.0;
        timeout_=0.0;
        unlock();
        for(vmp_index i=0;i<vclose.size();i++)
        {
            cell=vclose[i];
            cell->close_(cell);
            cell->release();
        }
        if(vread.size() > 0)
        {    
            if(loop_)
                select.fd_monitor(stimeout);
        }
        else
            break;
        lock();
        now_=timer_.now();
        unlock();
        for(vmp_index i=0;i<vread.size();i++)
        {    
            cell=vread[i];
            lock();
            vmp_bool rbool=cell->event_->active_;
            if((cell->timeout_) != 0.0 && (cell->timeout_<=now_))
            {    
                cell_close(cell,event::TIMEOUT);
                rbool=false;
            }
            pcell_.push(cell);
            unlock();
            if(rbool)
               cell->read_(cell);
        }
    }
    unlock();
}

}}

==> vmp-0.4.1.tar.xz: vmp-0.4.1/framework/src/event/manager.cc <==
/* -*- Mode:C++; c++-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Marco Guastella alias Vasta 
 * Web page:<www.ragnu.it> 
 * Email:<vasta@ragnu.it>
   Date last update: 15/05/2020
 */

#include "event.h"

namespace vampiria { namespace event {

Manager::Manager()
{
    timer_.init();
    self_=vmp::thread::pthread_self_wrap();
    init_=false;
    breakui_=new event::BreakloopUI(this);
    init();
}

Manager::~Manager()
{
    delete breakui_;
    breakui_=0;
}

void Manager::init()
{
    lock();
    if(!init_)
    {
        timeout_=0.0;
        now_=timer_.now();
        loop_=true;
        init_=true;
        breakui_->init();
        breakui_->cell_->priority_=event::BREAKLOOP_PRIORITY;
    }
    unlock();
}

void Manager::lock()
{
    mutex_.lock();
}

void Manager::unlock()
{
    mutex_.unlock();
}

void Manager::breakloop()
{
    if(self_ != vmp::thread::pthread_self_wrap())
       breakui_->active();       
}

event::Cell *Manager::cell_new(event::UI *ui,event::Event *event,vmp_int priority,event::EVTCB read,event::EVTCB close)
{
    if(!init_)
        vmp::except_s("event::Manager::cell_new() event Manager not initialized");
    event::Cell *cell=cellref_.get();
    cell->reset();
    cell->ui_=ui;
    cell->event_=event;
    event->cell_=cell;
    cell->manager_=this;
    cell->priority_=priority;
    cell->writing_=false;
    if(read == 0)
      cell->read_=event::empty_ev;
    else
      cell->read_=read;
    if(close == 0)
      cell->close_=event::empty_ev;
    else
      cell->close_=close;
    cell_alloc(cell);
    pcell_.push(cell);
    breakloop();
    return cell;  
}

vmp_bool Manager::cell_update(event::Cell *cell)
{
    if(cell->state_ == event::CLOSE)
       return false;
    if(cell->state_ == event::CLOSEWAIT)
    {
        if(cell->nwait_ == 0)
            return false;
        cell->nwait_--;
    }
    if(cell->timewait_ != 0.0)
       cell->timeout_=now_+cell->timewait_;
    return true;
}

void Manager::cell_timewait(event::Cell *cell,vmp::time::Time timewait)
{
    if(timewait > 0.0)
    {
        cell->timewait_=timewait;
        cell->timeout_=now_+cell->timewait_;
    }
    else
    {
        cell->timewait_=0.0;
        cell->timeout_=0.0;
    }
    breakloop();
}

void Manager::cell_close(event::Cell *cell,event::CELLRET retcode,vmp::str errinfo)
{
   if(cell->state_ != event::CLOSE)
   {
       cell->state_=event::CLOSE;
       cell->ret_=retcode;
       cell->err_=errinfo;
       cell->ui_->close_event(cell);
       breakloop();
   }
}

void Manager::cell_closewait(event::Cell *cell,vmp_uint n)
{
   if(cell->state_ == event::ACTIVE)
   {
       if(n == 0)
           cell_close(cell,event::SUCCESS);
       else
       {
           cell->state_=event::CLOSEWAIT;
           cell->nwait_=n;
       }
   }
}

vmp::time::Time Manager::time_now()
{
    return now_;
}

void Manager::set_timeout(vmp::time::Time timeout)
{
    if(timeout > 0.0)
    {
       if((timeout_ == 0.0) || (timeout < timeout_))
          timeout_=timeout;  
    }
}

void Manager::cell_alloc(event::Cell *cell)
{
    cell->refalloc_++;
}

void Manager::cell_release(event::Cell *cell)
{
    cell->refalloc_--;
    if(cell->refalloc_==0)
    {
        cell->ui_->free_ref(cell);
        cellref_.free(cell);
    }
}

void Manager::stop()
{
    loop_=false;
    breakui_->active();
}

void Manager::loop()
{
    if(!init_)
        vmp::except_s("event::Manager::loop() event Manager not initialized");
    register vmp::Select_FD select;
    register event::Cell *cell;
    register vmp::vector<event::Cell *> vread;
    register vmp::vector<event::Cell *> vclose;
    register vmp::time::Time stimeout;
    register vmp_size s;
    while(1)
    {
        lock();
        select.reset();
        vread.clear();
        vclose.clear();
        while(!pcell_.empty())
        {
            cell=pcell_.top();
            pcell_.pop();
            if(((cell->state_ == event::CLOSEWAIT) && (cell->nwait_ == 0)) || (!loop_))
                cell_close(cell,event::SUCCESS);   
            if(cell->state_ == event::CLOSE)
                vclose.push_back(cell);
            else
            {
                vread.push_back(cell);
                select.fd_add(cell->event_->fd_);
                if(cell->writing_)
                    select.fd_add_w(cell->event_->fd_);
            }
        }
        s=vclose.size();
        unlock();
        for(vmp_index i=0;i < s;i++)
        {
            cell=vclose[i];
            cell->close_(cell);
            cell->release();
        }
        lock();
        if(s > 0)
            breakui_->active();
        for(vmp_index i=0;i<vread.size();i++)
            set_timeout(vread[i]->timeout_);
        if(timeout_ > 0.0)
           stimeout=timeout_-now_;
        else
           stimeout=0.0;
        timeout_=0.0;
        unlock();
        if(vread.size() > 0)
        {    
            if(loop_)
                select.fd_monitor(stimeout);
        }
        else
            break;
        lock();
        now_=timer_.now();
        unlock();
        for(vmp_index i=0;i<vread.size();i++)
        {    
            cell=vread[i];
            lock();
            vmp_bool rbool=cell->event_->active_;
            if((cell->timeout_) != 0.0 && (cell->timeout_<=now_))
            {    
                cell_close(cell,event::TIMEOUT);
                rbool=false;
            }
            pcell_.push(cell);
            unlock();
            if(rbool)
               cell->read_(cell);
        }
    }
    init_=false;
    unlock();
}

}}
